"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at 

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import calendar
from threading import Thread
import logging
import time

from abc import abstractmethod, abstractproperty
from datetime import datetime as dt, datetime
from concurrent.futures.thread import ThreadPoolExecutor
from agora.client.agora import Agora, AGORA
from sdh.curator.server import app
from sdh.curator.store import r
from sdh.curator.store.triples import cache as cache, add_stream_triple, load_stream_triples, graph_provider
from sdh.curator.daemons.delivery import build_response
from redis.lock import Lock
from rdflib import RDF, RDFS
import random

__author__ = 'Fernando Serena'

log = logging.getLogger('sdh.curator.daemons.fragment')
agora_conf = app.config['AGORA']
agora_client = Agora('http://{}:{}'.format(agora_conf['host'], agora_conf['port']))

thp = ThreadPoolExecutor(max_workers=4)

fragment_locks = r.keys('*lock*')
for flk in fragment_locks:
    r.delete(flk)

fragment_pullings = r.keys('fragments:*:pulling')
for fpk in fragment_pullings:
    r.delete(fpk)

fragment_consumers = r.keys('fragments:*:consumers')
for fck in fragment_consumers:
    r.delete(fck)


class FragmentPlugin(object):
    __plugins = []

    @abstractmethod
    def consume(self, fid, quad, graph, *args):
        pass

    @abstractmethod
    def complete(self, fid, *args):
        pass

    @abstractproperty
    def sink_class(self):
        pass

    def sink_aware(self):
        return True

    @classmethod
    def register(cls, p):
        if issubclass(p, cls):
            cls.__plugins.append(p())
        else:
            raise ValueError('{} is not a valid fragment plugin'.format(p))

    @classmethod
    def plugins(cls):
        return cls.__plugins[:]


def __on_load_seed(s, _):
    log.debug('{} dereferenced'.format(s))


def __bind_prefixes(source_graph):
    map(lambda (prefix, uri): cache.bind(prefix, uri), source_graph.namespaces())


def map_variables(tp, mapping):
    if mapping is None:
        return tp
    return tuple(map(lambda x: mapping.get(x, x), tp))


def __consume_quad(fid, (c, s, p, o), graph, sinks=None):
    def __sink_consume():
        for rid in filter(lambda _: isinstance(sinks[_], plugin.sink_class), sinks):
            sink = sinks[rid]
            try:
                plugin.consume(fid, (map_variables(c, sink.mapping), s, p, o), graph, sink)
            except Exception as e:
                sink.remove()
                yield rid
                log.warning(e.message)

    def __generic_consume():
        try:
            plugin.consume(fid, (c, s, p, o), graph)
        except Exception as e:
            log.warning(e.message)

    for plugin in FragmentPlugin.plugins():
        if plugin.sink_class is not None:
            invalid_sinks = list(__sink_consume())
            for _ in invalid_sinks:
                del sinks[_]
        else:
            __generic_consume()


def __notify_completion(fid, sinks):
    for plugin in FragmentPlugin.plugins():
        try:
            filtered_sinks = filter(lambda _: isinstance(sinks[_], plugin.sink_class), sinks)
            for rid in filtered_sinks:
                sink = sinks[rid]
                if sink.delivery == 'accepted':
                    sink.delivery = 'ready'
                if plugin.sink_aware:
                    plugin.complete(fid, sink)
            if not plugin.sink_aware:
                plugin.complete(fid)
        except Exception as e:
            log.warning(e.message)


def __triple_pattern(graph, c):
    def extract_node_id(node):
        nid = node
        if (node, RDF.type, AGORA.Variable) in graph:
            nid = list(graph.objects(node, RDFS.label)).pop()
        elif (node, RDF.type, AGORA.Literal) in graph:
            nid = list(graph.objects(node, AGORA.value)).pop()
        return nid

    predicate = list(graph.objects(c, AGORA.predicate)).pop()
    subject_node = list(graph.objects(c, AGORA.subject)).pop()
    object_node = list(graph.objects(c, AGORA.object)).pop()
    subject = extract_node_id(subject_node)
    obj = extract_node_id(object_node)

    return str(subject), predicate.n3(graph.namespace_manager), str(obj)


def __replace_fragment(fid):
    tps = cache.get_context(fid).subjects(RDF.type, AGORA.TriplePattern)
    cache.remove_context(cache.get_context('/' + fid))
    for tp in tps:
        cache.remove_context(cache.get_context(str((fid, __triple_pattern(cache, tp)))))
    fragment_triples = load_stream_triples(fid, calendar.timegm(dt.now().timetuple()))
    for c, s, p, o in fragment_triples:
        cache.get_context(str((fid, c))).add((s, p, o))
        cache.get_context('/' + fid).add((s, p, o))
    with r.pipeline() as pipe:
        pipe.delete('fragments:{}:stream'.format(fid))
        pipe.execute()

def __cache_plan_context(fid, graph):
    try:
        cache.remove_context(cache.get_context(fid))
        fid_context = cache.get_context(fid)
        tps = graph.subjects(RDF.type, AGORA.TriplePattern)
        for tp in tps:
            for (s, p, o) in graph.triples((tp, None, None)):
                fid_context.add((s, p, o))
                for t in graph.triples((o, None, None)):
                    fid_context.add(t)
    except Exception, e:
        log.error(e.message)


def __pull_fragment(fid):
    def __load_fragment_requests():
        requests_ = r.smembers('fragments:{}:requests'.format(fid))
        sinks_ = {}
        for rid in requests_:
            try:
                sinks_[rid] = build_response(rid).sink
            except Exception, e:
                log.warning(e.message)
                with r.pipeline(transaction=True) as p:
                    p.multi()
                    p.srem('fragments:{}:requests'.format(fid), rid)
                    p.execute()
        return requests_, sinks_

    tps = r.smembers('fragments:{}:gp'.format(fid))
    requests, r_sinks = __load_fragment_requests()
    log.info('Pulling fragment {}, described by {}...'.format(fid, tps))
    fgm_gen, _, graph = agora_client.get_fragment_generator('{ %s }' % ' . '.join(tps), workers=2,
                                                            provider=graph_provider)

    triple_patterns = {tpn: __triple_pattern(graph, tpn) for tpn in
                       graph.subjects(RDF.type, AGORA.TriplePattern)}
    fragment_contexts = {tpn: (fid, triple_patterns[tpn]) for tpn in triple_patterns}
    __bind_prefixes(graph)

    lock_key = 'fragments:{}:lock'.format(fid)
    lock = r.lock(lock_key, lock_class=Lock)
    lock.acquire()

    lock_consume_key = 'fragments:{}:lock:consume'.format(fid)
    c_lock = r.lock(lock_consume_key, lock_class=Lock)
    c_lock.acquire()

    with r.pipeline(transaction=True) as p:
        p.multi()
        p.set('fragments:{}:pulling'.format(fid), True)
        p.delete('fragments:{}:contexts'.format(fid))
        for tpn in fragment_contexts.keys():
            p.sadd('fragments:{}:contexts'.format(fid), fragment_contexts[tpn])
        p.execute()
    lock.release()

    c_lock.release()

    n_triples = 0
    for (c, s, p, o) in fgm_gen:
        lock.acquire()
        if add_stream_triple(fid, triple_patterns[c], (s, p, o)):
            __consume_quad(fid, (triple_patterns[c], s, p, o), graph, sinks=r_sinks)
        lock.release()
        if r.scard('fragments:{}:requests'.format(fid)) != len(requests):
            requests, r_sinks = __load_fragment_requests()
        n_triples += 1

    log.debug('{} triples retrieved for fragment {}'.format(n_triples, fid))

    lock.acquire()
    c_lock.acquire()
    __replace_fragment(fid)
    __cache_plan_context(fid, graph)
    with r.pipeline(transaction=True) as p:
        p.multi()
        sync_key = 'fragments:{}:sync'.format(fid)
        p.set(sync_key, True)
        p.expire(sync_key, random.randint(60, 100))
        p.set('fragments:{}:updated'.format(fid), dt.now())
        p.delete('fragments:{}:pulling'.format(fid))
        p.execute()
    c_lock.release()
    __notify_completion(fid, r_sinks)
    lock.release()


def __collect_fragments():
    log.info('Collector thread started')

    futures = {}
    while True:
        for fid in filter(
                lambda x: r.get('fragments:{}:sync'.format(x)) is None and r.get(
                    'fragments:{}:pulling'.format(x)) is None,
                r.smembers('fragments')):
            if fid in futures:
                if futures[fid].done():
                    del futures[fid]
            if fid not in futures:
                futures[fid] = thp.submit(__pull_fragment, fid)
        time.sleep(1)


th = Thread(target=__collect_fragments)
th.daemon = True
th.start()



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at 

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import pkgutil
import logging
import os
import imp
import inspect
from sdh.curator.actions.core.base import Action

__author__ = 'Fernando Serena'

log = logging.getLogger('sdh.curator.actions')


def load_module(name):
    (importer, name, _) = action_modules[name]
    loader = importer.find_module(name)
    file_path = loader.get_filename()
    mod_name, file_ext = os.path.splitext(os.path.split(file_path)[-1])
    py_mod = None
    if file_ext.lower() == '.py':
        py_mod = imp.load_source(mod_name, file_path)
    elif file_ext.lower() == '.pyc':
        py_mod = imp.load_compiled(mod_name, file_path)

    action_modules[name] = py_mod

action_modules = {x[1]: x for x in pkgutil.iter_modules(path=['sdh/curator/actions/ext'])}
for module_name in action_modules:
    load_module(module_name)


def search_module(module, predicate, limit=1):
    py_mod = action_modules[module]

    if py_mod is not None:
        cand_elms = filter(predicate,
                           inspect.getmembers(py_mod, lambda x: inspect.isclass(x) and not inspect.isabstract(x)))
        if len(cand_elms) > limit:
            raise ValueError('Too many elements in module {}'.format(module))
        return cand_elms

    return None


def get_instance(module, clz, *args):
    module = action_modules[module]
    class_ = getattr(module, clz)
    instance = class_(*args)
    return instance


def execute(*args, **kwargs):
    log.debug('Searching for a compliant "{}" action handler...'.format(args[0]))
    name = args[0]

    try:
        _, clz = search_module(name,
                               lambda (_, cl): issubclass(cl, Action) and cl != Action).pop()
        data = kwargs.get('data', None)
        log.debug(
            'Found! Requesting an instance of {} to perform a/n {} action described as:\n{}'.format(clz, name, data))
        clz(data).submit()
    except IndexError:
        raise EnvironmentError('Action module found but class is missing: "{}"'.format(name))



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at 

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import logging
from abc import ABCMeta, abstractmethod, abstractproperty
from sdh.curator.actions.core.base import Request, Action, Response, Sink
from sdh.curator.actions.core.utils import CGraph
from rdflib import BNode, Literal, RDFS
from sdh.curator.actions.core import RDF, CURATOR, FOAF, TYPES, XSD
from sdh.curator.messaging.reply import reply
from sdh.curator.store import r
from datetime import datetime
import base64
import uuid

__author__ = 'Fernando Serena'

log = logging.getLogger('sdh.curator.actions.delivery')
CURATOR_UUID = Literal(str(uuid.uuid4()), datatype=TYPES.UUID)


def _build_reply_templates():
    accepted = CGraph()
    failure = CGraph()
    response_node = BNode()
    curator_node = BNode()
    accepted.add((response_node, RDF.type, CURATOR.Accepted))
    accepted.add((curator_node, RDF.type, FOAF.Agent))
    accepted.add((response_node, CURATOR.responseNumber, Literal("0", datatype=XSD.unsignedLong)))
    accepted.add((response_node, CURATOR.submittedBy, curator_node))
    accepted.add((response_node, CURATOR.submittedBy, curator_node))
    accepted.add(
        (curator_node, CURATOR.agentId, CURATOR_UUID))
    accepted.bind('types', TYPES)
    accepted.bind('curator', CURATOR)
    accepted.bind('foaf', FOAF)

    for triple in accepted:
        failure.add(triple)
    failure.set((response_node, RDF.type, CURATOR.Failure))
    for (prefix, ns) in accepted.namespaces():
        failure.bind(prefix, ns)

    return accepted, failure


def build_reply(template, reply_to, comment=None):
    reply_graph = CGraph()
    root_node = None
    for (s, p, o) in template:
        reply_graph.add((s, p, o))
        if p == RDF.type:
            root_node = s
    reply_graph.add((root_node, CURATOR.responseTo, Literal(reply_to, datatype=TYPES.UUID)))
    reply_graph.set((root_node, CURATOR.submittedOn, Literal(datetime.now())))
    reply_graph.set((root_node, CURATOR.messageId, Literal(str(uuid.uuid4()), datatype=TYPES.UUID)))
    if comment is not None:
        reply_graph.set((root_node, RDFS.comment, Literal(comment, datatype=XSD.string)))
    for (prefix, ns) in template.namespaces():
        reply_graph.bind(prefix, ns)
    return reply_graph

accepted_template, failure_template = _build_reply_templates()


class DeliveryRequest(Request):
    def __init__(self):
        super(DeliveryRequest, self).__init__()

    def _extract_content(self):
        super(DeliveryRequest, self)._extract_content()

        q_res = self._graph.query("""SELECT ?node ?ex ?rk ?h ?p ?v WHERE {
                                  ?node curator:replyTo [
                                        a curator:DeliveryChannel;
                                        amqp:exchangeName ?ex;
                                        amqp:routingKey ?rk;
                                        amqp:broker [
                                           a amqp:Broker;
                                           amqp:host ?h;
                                           amqp:port ?p;
                                           amqp:virtualHost ?v
                                        ]
                                     ]
                                  } """)
        q_res = list(q_res)
        if len(q_res) != 1:
            raise SyntaxError('Invalid delivery request')

        request_fields = q_res.pop()

        if not any(request_fields):
            raise ValueError('Missing fields for delivery request')

        if request_fields[0] != self._request_node:
            raise SyntaxError('Request node does not match')

        delivery_data = {}

        (delivery_data['exchange'],
         delivery_data['routing_key'],
         delivery_data['host'],
         delivery_data['port'],
         delivery_data['vhost']) = request_fields[1:]
        log.debug("""Parsed attributes of a delivery action request:
                      -exchange name: {}
                      -routing key: {}
                      -host: {}
                      -port: {}
                      -virtual host: {}""".format(
            delivery_data['exchange'],
            delivery_data['routing_key'],
            delivery_data['host'], delivery_data['port'], delivery_data['vhost']))

        self._fields['delivery'] = delivery_data.copy()

    @property
    def broker(self):
        broker_dict = {k: self._fields['delivery'][k].toPython() for k in ('host', 'port', 'vhost') if k in self._fields['delivery']}
        broker_dict['port'] = int(broker_dict['port'])
        return broker_dict

    @property
    def channel(self):
        return {k: self._fields['delivery'][k].toPython() for k in ('exchange', 'routing_key') if
                k in self._fields['delivery']}

    @property
    def recipient(self):
        recipient = self.broker.copy()
        recipient.update(self.channel)
        return recipient


class DeliveryAction(Action):
    __metaclass__ = ABCMeta

    def __init__(self, message):
        super(DeliveryAction, self).__init__(message)

    def __reply_accepted(self):
        graph = build_reply(accepted_template, self.request.message_id)
        log.debug('Notifying acceptance of request number {}'.format(self.request_id))
        reply(graph.serialize(format='turtle'), exchange='sdh',
              routing_key='curator.response.{}'.format(self.request.submitted_by),
              **self.request.broker)
        if self.sink.delivery is None:
            self.sink.delivery = 'accepted'

    def _reply_failure(self, reason=None):
        try:
            graph = build_reply(failure_template, self.request.message_id, reason)
            log.debug('Notifying failure of request number {}'.format(self.request_id))
            reply(graph.serialize(format='turtle'), exchange='sdh',
                  routing_key='curator.response.{}'.format(self.request.submitted_by),
                  **self.request.broker)
        except KeyError:
            # Delivery channel data may be invalid
            pass

    def submit(self):
        super(DeliveryAction, self).submit()
        try:
            self.__reply_accepted()
        except Exception, e:
            log.warning(e.message)
            self.sink.remove()
        # if self.sink.ready:
        #     self.sink.delivery = 'ready'


def used_channels():
    req_channel_keys = r.keys('requests:*')
    for rck in req_channel_keys:
        try:
            channel = r.hget(rck, 'channel')
            yield channel
        except Exception as e:
            log.warning(e.message)


def channel_sharing(channel_b64):
    return len(list(filter(lambda x: x == channel_b64, used_channels()))) - 1  # Don't count itself


class DeliverySink(Sink):
    __metaclass__ = ABCMeta

    @abstractmethod
    def _save(self, action):
        super(DeliverySink, self)._save(action)
        self._pipe.sadd('deliveries', self._request_id)
        broker_b64 = base64.b64encode('|'.join(map(lambda x: str(x), action.request.broker.values())))
        channel_b64 = base64.b64encode('|'.join(action.request.channel.values()))
        self._pipe.hmset('channels:{}'.format(channel_b64), action.request.channel)
        self._pipe.hmset('brokers:{}'.format(broker_b64), action.request.broker)
        self._pipe.hset('{}'.format(self._request_key), 'channel', channel_b64)
        self._pipe.hset('{}'.format(self._request_key), 'broker', broker_b64)

    @abstractmethod
    def _load(self):
        super(DeliverySink, self)._load()
        self._dict_fields['channel'] = r.hgetall('channels:{}'.format(self._dict_fields['channel']))
        self._dict_fields['broker'] = r.hgetall('brokers:{}'.format(self._dict_fields['broker']))
        self._dict_fields['broker']['port'] = int(self._dict_fields['broker']['port'])
        recipient = self._dict_fields['channel'].copy()
        recipient.update(self._dict_fields['broker'])
        self._dict_fields['recipient'] = recipient
        try:
            del self._dict_fields['delivery']
        except KeyError:
            pass

    @abstractmethod
    def _remove(self, pipe):
        super(DeliverySink, self)._remove(pipe)
        pipe.srem('deliveries', self._request_id)
        pipe.srem('deliveries:ready', self._request_id)
        channel_b64 = r.hget(self._request_key, 'channel')
        sharing = channel_sharing(channel_b64)
        if not sharing:
            log.debug('Removing channel {}'.format(channel_b64))
            pipe.delete('channels:{}'.format(channel_b64))

    @property
    def delivery(self):
        return r.hget('requests:{}'.format(self._request_id), 'delivery')

    @delivery.setter
    def delivery(self, value):
        with r.pipeline(transaction=True) as p:
            p.multi()
            if value == 'ready':
                p.sadd('deliveries:ready', self._request_id)
            elif value == 'sent':
                p.srem('deliveries:ready', self._request_id)
                p.sadd('deliveries:sent', self._request_id)
            p.hset('requests:{}'.format(self._request_id), 'delivery', value)
            p.execute()

    @abstractproperty
    def ready(self):
        return False


class DeliveryResponse(Response):
    __metaclass__ = ABCMeta

    def __init__(self, rid):
        super(DeliveryResponse, self).__init__(rid)

    @abstractmethod
    def build(self):
        super(DeliveryResponse, self).build()



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at 

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import calendar
import json

import logging
from datetime import datetime
import uuid

from sdh.curator.actions.core.fragment import FragmentRequest, FragmentAction, FragmentResponse, FragmentSink
from sdh.curator.actions.core import CURATOR, TYPES, RDF, XSD, FOAF
from sdh.curator.actions.core.utils import CGraph, chunks
from rdflib import BNode, Literal
from sdh.curator.actions.core.delivery import CURATOR_UUID
from sdh.curator.messaging.reply import reply
from sdh.curator.daemons.fragment import FragmentPlugin

__author__ = 'Fernando Serena'

log = logging.getLogger('sdh.curator.actions.query')


class QueryPlugin(FragmentPlugin):
    @property
    def sink_class(self):
        return QuerySink

    def consume(self, fid, (c, s, p, o), graph, *args):
        pass

    def complete(self, fid, *args):
        pass


FragmentPlugin.register(QueryPlugin)


class QueryRequest(FragmentRequest):
    def __init__(self):
        super(QueryRequest, self).__init__()

    def _extract_content(self):
        super(QueryRequest, self)._extract_content()

        q_res = self._graph.query("""SELECT ?node WHERE {
                                        ?node a curator:QueryRequest .
                                    }""")

        q_res = list(q_res)
        if len(q_res) != 1:
            raise SyntaxError('Invalid query request')

        request_fields = q_res.pop()
        if not all(request_fields):
            raise ValueError('Missing fields for query request')
        if request_fields[0] != self._request_node:
            raise SyntaxError('Request node does not match')


class QueryAction(FragmentAction):
    def __init__(self, message):
        self.__request = QueryRequest()
        self.__sink = QuerySink()
        super(QueryAction, self).__init__(message)

    @property
    def sink(self):
        return self.__sink

    @classmethod
    def response_class(cls):
        return QueryResponse

    @property
    def request(self):
        return self.__request

    def submit(self):
        try:
            super(QueryAction, self).submit()
        except Exception as e:
            log.debug('Bad request: {}'.format(e.message))
            self._reply_failure(e.message)


class QuerySink(FragmentSink):
    def _remove(self, pipe):
        super(QuerySink, self)._remove(pipe)

    def __init__(self):
        super(QuerySink, self).__init__()

    def _save(self, action):
        super(QuerySink, self)._save(action)
        if self.backed:
            self.delivery = 'ready'

    def _load(self):
        super(QuerySink, self)._load()


class QueryResponse(FragmentResponse):
    def __init__(self, rid):
        self.__sink = QuerySink()
        self.__sink.load(rid)
        super(QueryResponse, self).__init__(rid)

    @property
    def sink(self):
        return self.__sink

    def _build(self):
        fragment, _ = self.fragment(result_set=True)
        log.debug('Building a query result for request number {}'.format(self._request_id))

        try:
            variables = filter(lambda x: not x.startswith('_'), map(lambda v: v.lstrip('?'),
                                                                    filter(lambda x: x.startswith('?'),
                                                                           self.sink.preferred_labels)))

            for ch in chunks(fragment, 100):
                result_rows = []
                for t in ch:
                    result_row = {v: t[v] for v in variables}
                    result_rows.append(result_row)
                yield json.dumps(result_rows), {'state': 'streaming', 'source': 'store',
                                                'response_to': self.sink.message_id,
                                                'submitted_on': calendar.timegm(datetime.now().timetuple()),
                                                'submitted_by': self.sink.submitted_by}
        except Exception, e:
            log.error(e.message)
            raise
        finally:
            yield '', {'state': 'end'}



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at 

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

from rdflib import ConjunctiveGraph, URIRef, Literal, XSD
from sdh.curator.server import app
import logging
import calendar
from datetime import datetime as dt
from sdh.curator.store import r
import os
import shutil
import shortuuid
from concurrent.futures import ThreadPoolExecutor

__author__ = 'Fernando Serena'

log = logging.getLogger('sdh.curator.store.triples')
pool = ThreadPoolExecutor(max_workers=4)


def load_stream_triples(fid, until):
    def __triplify(x):
        def __term(elm):
            if elm.startswith('<'):
                return URIRef(elm.lstrip('<').rstrip('>'))
            elif '^^' in elm:
                (value, ty) = tuple(elm.split('^^'))
                return Literal(value.replace('"', ''), datatype=URIRef(ty.lstrip('<').rstrip('>')))
            else:
                return Literal(elm.replace('"', ''), datatype=XSD.string)

        c, s, p, o = eval(x)
        return c, __term(s), __term(p), __term(o)

    for x in r.zrangebyscore('fragments:{}:stream'.format(fid), '-inf', '{}'.format(float(until))):
        yield __triplify(x)


def add_stream_triple(fid, tp, (s, p, o), timestamp=None):
    if timestamp is None:
        timestamp = calendar.timegm(dt.utcnow().timetuple())
    quad = (tp, s.n3(), p.n3(), o.n3())
    stream_key = 'fragments:{}:stream'.format(fid)
    not_found = not bool(r.zscore(stream_key, quad))
    if not_found:
        with r.pipeline() as pipe:
            pipe.zadd(stream_key, timestamp, quad)
            pipe.execute()
    return not_found


class GraphProvider(object):

    def __init__(self):
        self.__graph_dict = {}

    @staticmethod
    def __clean(name):
        shutil.rmtree('store/query/{}'.format(name))

    def create(self, conjunctive=False):
        uuid = shortuuid.uuid()
        if conjunctive:
            if 'persist' in app.config['STORE']:
                g = ConjunctiveGraph('Sleepycat')
                g.open('store/query/{}'.format(uuid), create=True)
                g.store.graph_aware = False
            else:
                g = ConjunctiveGraph()
                g.store.graph_aware = False
        else:
            g = query.get_context(uuid)
        self.__graph_dict[g] = uuid
        return g

    def destroy(self, g):
        if g in self.__graph_dict:
            if isinstance(g, ConjunctiveGraph):
                if 'persist' in app.config['STORE']:
                    g.close()
                    pool.submit(self.__clean, self.__graph_dict[g])
                else:
                    g.remove((None, None, None))
                    g.close()
            else:
                query.remove_context(query.get_context(self.__graph_dict[g]))
            del self.__graph_dict[g]


graph_provider = GraphProvider()
store_mode = app.config['STORE']
if 'persist' in store_mode:
    if not os.path.exists('store'):
        os.makedirs('store')
    if os.path.exists('store/query'):
        shutil.rmtree('store/query/')
    if os.path.exists('store/tmp'):
        shutil.rmtree('store/tmp')
    os.makedirs('store/query')
    log.info('Loading known triples...')
    cache = ConjunctiveGraph('Sleepycat')
    cache.open('store/cache', create=True)
    query = ConjunctiveGraph('Sleepycat')
    query.open('store/tmp', create=True)
else:
    cache = ConjunctiveGraph()
    query = ConjunctiveGraph()

cache.store.graph_aware = False
query.store.graph_aware = False
log.info('Ready')



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  This file is part of the Smart Developer Hub Project:
    http://www.smartdeveloperhub.org

  Center for Open Middleware
        http://www.centeropenmiddleware.com/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2015 Center for Open Middleware.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at 

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

from rdflib import Graph
import itertools
import networkx as nx
from networkx.algorithms.isomorphism import DiGraphMatcher

__author__ = 'Fernando Serena'


def parse_bool(s):
    if type(s) == str:
        if s == 'True':
            return True
    return False


class CGraph(Graph):
    def objects(self, subject=None, predicate=None, card=None):
        objs_gen = super(CGraph, self).objects(subject, predicate)
        if card is None:
            return objs_gen

        objs_gen, gen_bak = itertools.tee(objs_gen)
        objs = list(objs_gen)

        if card == 1:
            if not (0 < len(objs) < 2):
                raise ValueError(len(objs))
            return objs.pop()

        return gen_bak

    def subjects(self, predicate=None, object=None, card=None):
        subs_gen = super(CGraph, self).subjects(predicate, object)
        if card is None:
            return subs_gen

        subs_gen, gen_bak = itertools.tee(subs_gen)
        subs = list(subs_gen)

        if card == 1:
            if not (0 < len(subs) < 2):
                raise ValueError(len(subs))
            return subs.pop()

        return gen_bak


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    if n:
        if getattr(l, '__iter__') is not None:
            l = l.__iter__()
        finished = False
        while not finished:
            chunk = []
            try:
                for _ in range(n):
                    chunk.append(l.next())
            except StopIteration:
                finished = True
            yield chunk


class GraphPattern(set):
    def __init__(self, s=()):
        super(GraphPattern, self).__init__(s)

    @property
    def gp(self):
        return self

    @property
    def wire(self):
        g = nx.DiGraph()
        for tp in self:
            (s, p, o) = tuple(tp.split(' '))
            edge_data = {'link': p}
            g.add_node(s)
            if o.startswith('?'):
                g.add_node(o)
            else:
                g.add_node(o, literal=o)
                edge_data['to'] = o
            g.add_edge(s, o, **edge_data)

        return g

    def __eq__(self, other):
        if not isinstance(other, GraphPattern):
            return super(GraphPattern, self).__eq__(other)

        mapping = self.mapping(other)
        return mapping is not None

    def mapping(self, other):
        if not isinstance(other, GraphPattern):
            return None

        my_wire = self.wire
        others_wire = other.wire

        def __node_match(n1, n2):
            return n1 == n2

        def __edge_match(e1, e2):
            return e1 == e2

        matcher = DiGraphMatcher(my_wire, others_wire, node_match=__node_match, edge_match=__edge_match)
        mapping = list(matcher.isomorphisms_iter())
        if len(mapping) == 1:
            return mapping.pop()
        else:
            return None






